##  Using BEAGLE R template, version of 28/07/2016 :
##  rule written by plum.py ;
##  derived from training data : C:\beagling\op\elements_dat1.dat;
##  generated on creation date : Fri Aug 12 17:18:51 2016;
##  dumped on Fri Aug 12 17:22:56 2016.
##
beag_gold = 5.0 ** 0.5 * 0.5 + 0.5  ##  global

##  helper functions :
beag_exor = function (v1,v2) {
	##  exclusive or, as in Beagle :
	return ((v1>0) != (v2>0))
	}

beag_root = function (v) {
	##  safe square root :
	if (v >= 0.0)  return (sqrt(v))
	else  return (-sqrt(abs(v)))
	}
beag_slog = function (v) {
	##  safe natural logarithm :
	if (v < 0)  return (-log(1+abs(v)))
	else  return  (log(1+v))
	}

beag_stabprep = function () {
	##  sets up fallout table :

	stab = list()
	stab[['000']] = c(1, 0, 0, 0)
	stab[['001']] = c(0, 0, 0, 2)
	stab[['010']] = c(16, 0, 0, 0)
	stab[['011']] = c(0, 0, 27, 0)
	stab[['100']] = c(0, 0, 0, 0)
	stab[['101']] = c(0, 0, 0, 10)
	stab[['110']] = c(4, 0, 0, 0)
	stab[['111']] = c(0, 20, 0, 0)

	##  unpacks stab lines.

	return (stab)
	}  ##  stabprep ends.


beag_decrule = function (vals,stab) {
	##  input vals should be a 1-row dataframe with appropriate colnames.
	##  target : block
	##  rule mode is tabular.

	rule = c() ; bins = c('0','1')
	catlist = c('d', 'f', 'p', 's')
	priorvec = c(0.25868385949580663, 0.25244959735256584, 0.2933199258909766, 0.19554661726065106)
	subrules = 3
	##  compute rule values :
	rule[1] = (vals$grouping < vals$orbitals)
	rule[2] = (4.0758038 < (vals$atweight * (vals$grouping > vals$outshell)))
	rule[3] = (grepl("Transition metal",vals$type)<= 0)

	p = 0 ; b = c()
	while (p < subrules) {
		p = p + 1  ##  early-r, late-py
		v = (rule[p]>0)  ##  omit if demonic
		b = c(b,bins[v+1])  ##  boolean string, omit if demonic
		}
	b = paste(b,collapse='')  ##  omit if demonic
	##  retrieve cell frequencies :
	frex = stab[[b]]
	cellsize = sum(frex)
	##  attenuate frex :
	cats = length(priorvec)
	slug = beag_gold / cats
	modfrex = slug + frex
	##  defer to Reverend Bayes :
	postvec = priorvec * modfrex
	postvec = postvec / sum(postvec)
	pc = which.max(postvec)
	predcat = catlist[pc]

	return (list(cellcode=b,predcat=predcat,frex=frex,postvec=postvec))
	}
##  decision rule ends.


leaflet = function (datframe) {
	##  Language-Export Application For Likelihood Estimation Trials.

	stab = beag_stabprep()
	rows = dim(datframe)[1]
	if (rows < 1)  return (NULL)
	options(stringsAsFactors=FALSE)
	##  prepare extra cols :
	predcat = character(rows) ; stabcode = character(rows)
	cellsize = numeric(rows)
	strength = numeric(rows)
	topprob = numeric(rows) ; nextprob = numeric(rows)
	for (r in 1:rows) {
		beagvals = beag_decrule(datframe[r,],stab)
		if (r == 1) cats = length(beagvals$frex)
			##  what if cats < 2 (?!)
			##  also test if predvals[[2]] is character ?
		stabcode[r] = beagvals$cellcode
		predcat[r] = as.character(beagvals$predcat)  ##  might be numeric
		cellsize[r] = sum(beagvals$frex)
		post = sort(beagvals$postvec,decreasing=TRUE)
		topprob[r] = post[1] ; nextprob[r] = post[2]
		##  nextprob's category tricky to obtain; worth doing ?
		pd = post[1] - post[2]  ##  winning margin
		if (cats <= 2)  strength[r] = pd
		else {
			ps = c(nextprob[r],post[2:cats])  ##  duplicate second-highest prob
			strength[r] = topprob[r] - sum(ps) / cats  ##  compare top with augmented remainder
			}
		}
	outframe = data.frame(strength,predcat,topprob,nextprob,stabcode,cellsize)
	temp = cbind(datframe,outframe)
	options(stringsAsFactors=default.stringsAsFactors())
	return (temp)
	}  ##  returns data frame with additional columns.

##  ending.
