"Python3 basic evolutionary library, for BEAGLE, by R.S. Forsyth."

##  EVOSOFT :
##
##  Python3 main evolutionary library :
##  Copyright 2016, R.S. Forsyth.
##  Released under GNU Public licence, version 3:
##  http://www.gnu.org/licenses/gpl.txt
##  first version : 30/11/2015
##  last revision : 12/03/2016


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
    This copyrighted program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)
    
	
import  math, os, string, sys
import  copy
import  random


from globdefs import *
from glib import  showobj


Showtop = 1  ##  1 = show new best scores
Evopars = Sack("default parameters")
Evopars.m1 = 0.5  ##  mutation prob, post-mating
Evopars.m2 = 0.25 ##  mutation prob, as separate operator
Evopars.ntrials = 8192
Evopars.popsize = 144
Evopars.maxdepth = 5  ## won't go into rulelib !
Evopars.probings = 4  ##  zero or 1 means 'duel' selection
##  functions that need to be passed in:
##  mate, clobber, makerule, (evalrule?), fitness, copyrule, (showrule?).
Funclist = ["mate","clobber","makerule","evalrule","fitness","copyrule"]


##  EVOPREP :
def evoprep (evopars):
    "sets up internal parameters passed thru object evopars"

    pars = dir(evopars)
    for a in pars:  ##  bring all attributes of input argument
        if a.startswith("__"):
            continue  ##  private attribute
        v = getattr(evopars,a)
        ##  print(a,v)
        setattr(Evopars,a,v)
        if a in Funclist and (Showtop > 1):
            print("function",a,"defined.")

    if not hasattr(Evopars,"fitness"):
        print("Warning: no fitness function provided to evoprep.")

    if Showtop > 1:
        showobj(Evopars)
        ##  print(dir(Evopars))

    return  ##  global Evopars altered.


##  EVOTRIES :
def evotries (context,poplist=[],evopars=Evopars):
    "evolves rules for cases with indices in training."
    ##  evolutionary context is a Sack(), usually paradat.

    if not hasattr(evopars,"fitness"):
        raise RuntimeError("No fitness function defined!")
    else:
        fitfunc = evopars.fitness
    ##  initipop normally called by calling program :
    if not poplist:
        poplist = initpop(context,evopars)
    topscore = fitfunc(context,poplist[0])
    bestpos = 0
    popsize = evopars.popsize
    assert popsize == len(poplist)
    for gp in range(1,popsize):
        r = poplist[gp]
        f = fitfunc(context,r)
        r.quality = f  ##  keep fitness score
        r.whengen = gp
        if f > topscore:  ##  always maximizing ?
            topscore = f
            bestpos = gp
            bestitem = evopars.copyrule(r)
    if Showtop > 0:
        ##  show initial high score :
        print(popsize,topscore)
    bestitem = evopars.copyrule(poplist[bestpos])  ##  redundant
    loop = len(poplist)
    evopars.lo = 0 ; evopars.hi = popsize - 1
    ##  main loop :
    while loop < evopars.ntrials:
        if evopars.probings > 1:
            p0,p1 = probing(context,poplist)
        else:  ##  default is 'duel' selection :
            p0 = random.randint(evopars.lo,evopars.hi)  ##  2b deceased
            p1 = random.randint(evopars.lo,evopars.hi)  ##  parent 1
            if poplist[p0].quality > poplist[p1].quality:  ##  reorder
                p0,p1 = p1,p0  ##  swap
        p2 = random.randint(evopars.lo,evopars.hi)  ## parent 2
        evopars.mate(context,poplist,p2,p1,p0)  ##  replaces p0
        loop += 1
        ##  print("crossed",loop)
        ##  showobj(poplist[p0])
        r = random.random()
        if r <= evopars.m1:  ##  mutation
            evopars.clobber(context,poplist,p0)
        ##  tidying required here ?
        f = fitfunc(context,poplist[p0])
        try:
            poplist[p0].quality = f  ##  associate fitness score with item
        except AttributeError as xxxx:
            print(xxxx)
            print(p0,f)
            print(poplist[p0])
        poplist[p0].whengen = loop
        if f > topscore:  ##  save in archive
            topscore = f
            bestpos = p0
            bestitem = evopars.copyrule(poplist[p0])
            if Showtop:
                print(loop+1,f)
        ##  sometimes random mutation as well :
        if random.random() <= evopars.m2:
            p = random.randint(evopars.lo,evopars.hi)
            evopars.clobber(context,poplist,p)
            loop += 1
            poplist[p].quality = fitfunc(context,poplist[p])
            poplist[p].whengen = loop
            if poplist[p].quality > topscore:
                topscore = poplist[p].quality
                bestpos = p
                bestitem = evopars.copyrule(poplist[p])
                if Showtop:
                    print(loop+1,topscore)

    return  bestitem


##  INITPOP :
def initpop (context,evopars=Evopars):
    "makes random initial population of rules."

    popsize = evopars.popsize
    poplist = []
    for gp in range(popsize):
        r = evopars.makerule(context)
        ##  r.quality = evopars.fitfunc(r)  ## quality computed outside
        poplist.append(r)

    return  poplist


##  PROBING :
def probing (context,poplist,evopars=Evopars):
    "probes for mini & maxi scores."

    fitlist = []
    for p in range(evopars.probings):
        rp = random.randint(evopars.lo,evopars.hi)
        fitlist.append((poplist[rp].quality,rp))
    hi = max(fitlist)
    lo = min(fitlist)

    return  lo[1], hi[1]  ##  last item is position

##  fairly minimal, just a generic skeleton.

