"Python3 rule handling library, Copyright R.S. Forsyth."

##  RULELIB :
##
##  Python3 rule-handling library, for BEAGLE & RUNSTER :
##  Copyright 2016, R.S. Forsyth.
##  Released under GNU Public licence, version 3:
##  http://www.gnu.org/licenses/gpl.txt
##  first version : 30/11/2015
##  last revision : 02/09/2016


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
    This copyrighted program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)
    
	
import  math, os, string, sys
import  copy
import  random
import  time


from globdefs import *
from glib  import  showobj
from glib  import  numeric, numerize
import  treelib


Showoff = 1  ##  display control parameter
Star = '*'
Buck = '$'


##  functions needed :
##  mate, zonk, makerule, evalrule, fitness, copyrule, (ruleshow).


##  MAKERULE :
def makerule (paradat,full=True):
    "makes a fresh rule object with random values."

    r = Sack(paradat.traindat)  ##  id is training data
    r.quality = [-Huge]  ##  will be lexicographic
    r.stab = {}  ##  dict, indexed by binstring
    ##  r.testlist = []  ##  list of tests
    if full:  ##  normally make random expressions
        r.testlist = []  ##  complete refresh
        for t in range(paradat.subrules):
            test = treelib.maketree(paradat,random.randint(2,9))
            r.testlist.append(test)
    r.secs = time.time()
    ##  .dateline from paradat when required ?

    return  r
    ##  does stab need to be emptied ?


##  MATE :
def mate (paradat,rulepop,p2,p1,p0):
    "replaces rulepop[p0] by combining p1 & p2."

    rulepop[p0].quality = [-Huge]
    rulepop[p0].stab = {}
    try:
        nt = min(len(rulepop[p1].testlist),len(rulepop[p2].testlist))
    except AttributeError as xxxx:
        print(p0,p1,p2)
        rout = ruleshow(paradat,rulepop[p1])
        print(rout)
        ##  rout = ruleshow(paradat,rulepop[p2])
        ##  print(rout)
        showobj(rulepop[p2])
    assert nt == paradat.subrules
    m = rulepop[p2] ; p = rulepop[p1]
    ##  at present, number of tests can't differ.
    for t in range(nt):
        r = random.randint(1,4)
        try:
            if r == 1:  ##  paternal
                rulepop[p0].testlist[t] = treelib.copytree(p.testlist[t])
            elif r == 4:  ##  maternal
                rulepop[p0].testlist[t] = treelib.copytree(m.testlist[t])
            else:  ##  combined
                babe = treelib.mate(paradat,m.testlist[t],p.testlist[t])
                rulepop[p0].testlist[t] = babe
        except IndexError as xxxx:
            print(p0,p1,p2,t)
            raise RuntimeError(xxxx)
            
    ##  wd hav2 deal with unequal lengths here.

    return


##  ZONK :
def zonk (paradat,rulepop,p):
    "mutates rulepop[p]."

    nt = len(rulepop[p].testlist)
    z = random.randint(0,nt-1)
    rx = random.random()
    ##  rulepop[p] = makerule(paradat,False)  ##  clear fields
    rulepop[p].quality = [-Huge]
    rulepop[p].stab = {}
    if rx > 0.8:  ##  complete replacement
        rulepop[p] = makerule(paradat)
    else:  ##  alter single element
        rulepop[p].testlist[z] = treelib.zonk(paradat,rulepop[p].testlist[z])

    ##  print("zonk exit:",p)
    ##  print(type(rulepop[p]))
    ##  showobj(rulepop[p])
    return


##  INT2BIN :
def int2bin (i,wide):
    "converts integer to bitstring."

    b = bin(i)[2:]  ##  first 2 chars always "0b".

    return  b.rjust(wide,'0')


##  LIST2BIN :
def list2bin (booleans):
    "converts list of logical values to bitstring."

    if not booleans:
        return  ''  ##  '0' ??
    z = ['0'] * len(booleans)
    for j,bv in enumerate(booleans):
        if (bv+0) > 0:
            z[j] = '1'

    return  ''.join(z)


##  RULESHOW :
def ruleshow (paradat,r,catlist=[],priorvec=[]):
    "makes printable version of a rule as a line list."

    outlines = [r.id]
    if not hasattr(r,"dateline"):
        r.dateline = paradat.dateline
    if not hasattr(r,"rulemode"):
        r.rulemode = paradat.rulemode
    outlines.append(r.dateline)
    outlines.append(r.rulemode)
    outlines.append(str(paradat.rows) + ' ' + str(paradat.cols))
    ##  could be held internally ?
    if not hasattr(r,"targval"):
        r.targval = paradat.targval
    if hasattr(r,"catlist"):
        pass  ##  already present
    elif catlist:  ##  in regression mode catlist must be supplied [targcuts]
        r.catlist = catlist
    else:
        r.catlist = paradat.catlist  ##  categorical
    if hasattr(r,"priorvec"):
        pass  ##  already incorporated
    elif priorvec:
        r.priorvec = priorvec
    else:  ##  presumably in classification mode
        ##  r.priorvec = [round(p,9) for p in paradat.priorvec]
        r.priorvec = paradat.priorvec
    outlines.append(r.targval)
    outlines.append(r.catlist)
    outlines.append(r.priorvec)
    outlines.append(Buck)
    for test in r.testlist:
        outlines.append(treelib.tree2str(test,paradat))
    outlines.append(Buck)
    tabsize = paradat.stabrows
    for j in range(tabsize):
        if paradat.rulemode == "demonic":
            b = str(j)
        else:
            b = int2bin(j,paradat.subrules)
        if b in r.stab:
            trow = r.stab[b]  ##  fallout table
        else:
            trow = []
        outlines.append(b+"  "+str(trow))
    outlines.extend([str(r.quality),Buck])
    ##  evenually nwill have to dump taglist (used tags with positions?)

    return  outlines


##  RULEDUMP :
def ruledump (paradat,rout,fo):
    "dumps single rule-outlist onto opened file fo."

    print("training data :",rout[0],file=fo)
    ruledate = time.asctime(time.localtime(time.time()))
    print("creation date :",ruledate,file=fo)
    print("rule mode :",rout[2],file=fo)
    for outline in rout[3:]:
        print(outline,file=fo)

    return


##  GETRULE :
def getrule (paradat,datlines):
    "returns rulesack from lines read from rule file."

    rule = Sack('')
    nd = len(datlines)
    assert nd > 10  ##  fail gracefully later !
    rule.id = datlines[0].strip()
    rule.dateline = datlines[1].strip()
    rm = datlines[2].strip()
    rm = rm.split()
    rule.rulemode = rm[-1]  ##  last word is mode itself
    r,c = datlines[3].strip().split()
    rule.rows = int(r) ; rule.cols = int(c)
    rule.targval = datlines[4].strip()
    rule.catlist = eval(datlines[5])
    rule.priorvec = eval(datlines[6])
    p = 8  ##  p = 7
    rule.testlist = []
    while p < nd:
        datline = datlines[p].strip()
        p += 1
        if datline == Buck:  ##  section breaker
            break
        item = treelib.str2tree(paradat,datline)
        rule.testlist.append(item)
    ##  read fallout table :
    rule.stab = {}
    while p < nd:
        datline = datlines[p].strip()
        p += 1
        if datline == Buck:
            break  ##  error ??
        if datline.startswith('['):  ##  fallout table finished
            rule.quality=eval(datline)
            break
        bits = datline.split(maxsplit=1)
        ##  print(datline,bits)
        stabcell = bits[0]
        rule.stab[stabcell] = eval(bits[1])      
    ##  .stabrows & .subrules need 2b reset ?
    rule.subrules = len(rule.testlist)
    rule.stabrows = len(rule.stab)
        
    return  rule
    ##  should be as dumped by dumprule, plus inactive fields.


##  COPYRULE :
def copyrule (r):
    "makes fresh copy of rule object r."

    rc = copy.deepcopy(r)
    ##  if that fails, will need a dedicated function (& copytest?).

    return  rc


##  RULE2SIG :
def rule2sig (paradat,thisrule,rowvals):
    "evaluates testlist of rule to produce signature string."

    if paradat.rulemode == "demonic":
        return  rule2pos(paradat,thisrule,rowvals)
    ##  else default is tabular :
    vals = []
    for test in thisrule.testlist:
        v = treelib.treeval(test,paradat,rowvals)
        vals.append((v+0))  ##  force f/t to 0/1

    return  list2bin(vals)


##  RULE2POS :
def rule2pos (paradat,thisrule,rowvals):
    "gives sigstring as position of maximum of rule vals."

    assert paradat.subrules == len(thisrule.testlist)
    vals = [treelib.treeval(thisrule.testlist[j],paradat,rowvals) \
            for j in range(paradat.subrules)]
    vmax = max(vals)
    vp = vals.index(vmax)
    
    return  str(vp)    


##  STABFILL :
def stabfill (paradat,thisrule):
    "fills signature table from (training) cases."

    indices = paradat.training  ##  case indices given thru paradat
    ##  ensure empty sigtable :
    ##  w = len(thisrule.testlist)
    ##  for j in range(paradat.subrules):
        ##  stabcell = int2bin(j,w)
        ##  thisrule.stab[stabcell] = []
    ##  clear sigtable :
    thisrule.stab = {}
    for j in indices:
        rowvals = paradat.datamat[j]
        stabcell = rule2sig(paradat,thisrule,rowvals)
        ##  targ = rowvals[paradat.targcol]
        targ = paradat.targvals[j]
        tpos = paradat.catlist.index(targ)
        if not (stabcell in thisrule.stab) or \
           thisrule.stab[stabcell] == []:  ##  first encounter
            thisrule.stab[stabcell] = [0] * paradat.cats
        thisrule.stab[stabcell][tpos] += 1

    return


##  EVALRULE :
def evalrule (paradat,thisrule,rowvals):
    "evaluates single classification rule on single case."

    stabcell = rule2sig(paradat,thisrule,rowvals)
    if stabcell in thisrule.stab:
        frex = thisrule.stab[stabcell]
        if not frex:
            frex = [0] * paradat.cats
    else:
        frex = [0] * paradat.cats
    
    return  stabcell,frex
    ##  just returns cell frex, posteriors 2b computed outside.


##  EVALREGS :
def evalregs (paradat,thisrule,rowvals):
    "evaluates regression rule on single case."

    if paradat.rulemode != "tabular":
        v = treelib.treeval(thisrule.testlist[0],paradat,rowvals)
        return  ['0',[v]]
    ##  default is standard mode :
    stabcell = rule2sig(paradat,thisrule,rowvals)
    if stabcell in thisrule.stab:  ##  always present ?
        stat = thisrule.stab[stabcell]
    else:  ##  fallback
        stat = [1,paradat.targnorm,paradat.targstat.std1]

    return  stabcell,stat


##  FITNESS :
def fitness (paradat,thisrule):
    "fitness function for rule structures."
    ##  only valid with tabular rule mode.
    
    if not hasattr(thisrule,"stab"):  ##  sigtable not built !
        return  0.0  ##  negative as error signal ??
    ##  sigtable needs 2b filled :
    stabfill(paradat,thisrule)
    cattots = [0] * paradat.cats
    okay = 0
    freqvec = [0] * paradat.stabrows
    fi = 0
    for stabcell in thisrule.stab:
        frex = thisrule.stab[stabcell]
        freqvec[fi] = sum(frex) ; fi += 1
        if frex:
            okay += max(frex)
            for j in range(paradat.cats):
                cattots[j] += frex[j]
    n = sum(cattots)
    mc = max(cattots)

    if (n < 1) or (n == mc):
        gklambda = 0.0
    else:
        gklambda = (okay - mc) / (n - mc)

    ##  compare to profile :
    ##  freqvec.sort(reverse=True)
    ##  d = 0.0 ; n1 = 1.0 / n
    ##  e = 1.0 / paradat.stabrows
    ##  for fi in range(paradat.stabrows):
            ##  d += abs(paradat.profile[fi]-freqvec[fi]*n1)
        ##  d += abs(e-freqvec[fi]*n1)
        
    if paradat.brevity == 0:
        return  [gklambda]
    ##  otherwise append size penalty :
    totsize = 0
    for t in thisrule.testlist:
        totsize += treelib.size(t)

    return  [gklambda,-totsize/len(thisrule.testlist)]
    ##  inefficient to keep recomputing cattots.


##  DEMONFIT :
def demonfit (paradat,thisrule):
    "fitness function for rule structures, demonic mode."
    
    if not hasattr(thisrule,"stab"):  ##  sigtable not built !
        return  [0.0]  ##  negative as error signal ??
    ##  sigtable needs 2b filled :
    stabfill(paradat,thisrule)
    correct = 0 ; n = 0
    for p in range(paradat.subrules):
        ##  tc = paradat.catlist[p]
        try:
            frex = thisrule.stab[str(p)]
        except KeyError as xxxx:
            ##  print(p)
            ##  showobj(thisrule)
            ##  print(xxxx)
            ##  assert xxxx == 0  ##  forced ending
            frex = [0] * paradat.cats
        correct += frex[p]
        n += sum(frex)
    ##  proportional reduction in error :
    prie = (correct/n - paradat.guessing) \
           / (1.0 - paradat.guessing)

    if paradat.brevity == 0:
        return  [prie]
    ##  else append size penalty :
    totsize = 0
    for t in thisrule.testlist:
        totsize += treelib.size(t)

    return  [prie,-totsize/len(thisrule.testlist)]


##  REGSCORE :
def regscore (paradat,thisrule):
    "fitness function for simple symbolic regression."

    tote = 0.0  ##  total absolute error
    for j in paradat.training:
        rowvals = paradat.datamat[j]
        ##  tv = rowvals[paradat.targcol]
        tv = paradat.targvals[j]
        v = treelib.treeval(thisrule.testlist[0],paradat,rowvals)
        tote += abs(v - tv)
    d = tote / paradat.rows  ##  mean absolute deviation
    
    if paradat.brevity == 0:
        return  [-d]
    ##  else append size penalty :
    s = treelib.size(thisrule.testlist[0])

    return  [-d,-s]  ##  both divergences


##  REGFILL :
def regfill (paradat,thisrule):
    "fills signature table from (training) cases for tabular regression."

    indices = paradat.training  ##  case indices given thru paradat
    ##  clear sigtable :
    thisrule.stab = {}
    for j in indices:
        rowvals = paradat.datamat[j]
        stabcell = rule2sig(paradat,thisrule,rowvals)
        ##  targ = rowvals[paradat.targcol]
        targ = paradat.targvals[j]
        t2 = targ*targ
        if not (stabcell in thisrule.stab) or \
           thisrule.stab[stabcell] == []:  ##  first encounter
            ##  initial sums & sqrsums :
            thisrule.stab[stabcell] = [1,paradat.targnorm,
                                       paradat.targnorm**2]
        thisrule.stab[stabcell][0] += 1
        thisrule.stab[stabcell][1] += targ
        thisrule.stab[stabcell][2] += t2
        
    s1 = paradat.targstat.std1
    w = paradat.subrules
    for sr in range(paradat.stabrows):
        stabcell = int2bin(sr,w)
        if not(stabcell in thisrule.stab):  ##  fallback
            thisrule.stab[stabcell] = [1,paradat.targnorm,s1]
        else:
            nc = thisrule.stab[stabcell][0]
            xbar = thisrule.stab[stabcell][1] / nc
            v = (thisrule.stab[stabcell][2] / nc) - (xbar*xbar)  ##  variance
            ##  revise :
            thisrule.stab[stabcell][1] = xbar
            thisrule.stab[stabcell][2] = abs(v) ** 0.5  ##  sd
            
    return


##  QUALITAB :
def qualitab (paradat,thisrule):
    "fitness function for rule structures, tabular regression mode."
    ##  only valid with tabular-regression rule mode.
    
    if not hasattr(thisrule,"stab"):  ##  sigtable not built !
        return  0.0  ##  negative as error signal ??
    ##  sigtable needs 2b filled :
    regfill(paradat,thisrule)

    tote = 0.0  ##  total absolute error
    for j in paradat.training:
        rowvals = paradat.datamat[j]
        ##  tv = rowvals[paradat.targcol]
        tv = paradat.targvals[j]
        stabcell = rule2sig(paradat,thisrule,rowvals)
        v = thisrule.stab[stabcell][1]  ##  item 1 is mean
        tote += abs(v - tv)
    d = tote / paradat.rows  ##  mean absolute deviation
    
    if paradat.brevity == 0:
        return  [-d]
    ##  otherwise append size penalty :
    totsize = 0
    for t in thisrule.testlist:
        totsize += treelib.size(t)

    return  [-d,-totsize/len(thisrule.testlist)]


##  QUICKFIT :
def quickfit (paradat,thisrule):
    "quick & dirty fitness function for tabular regression."

    indices = paradat.training  ##  case indices given thru paradat
    ##  clear sigtable :
    thisrule.stab = {}
    targnorm = paradat.targnorm  ##  better as actual mean ??
    for j in indices:
        rowvals = paradat.datamat[j]
        stabcell = rule2sig(paradat,thisrule,rowvals)
        ##  targ = rowvals[paradat.targcol]
        targ = paradat.targvals[j]
        tdif = targ - targnorm
        if not (stabcell in thisrule.stab) or \
           thisrule.stab[stabcell] == []:  ##  first encounter
            ##  initial total :
            thisrule.stab[stabcell] = [1,tdif]
        else:
            thisrule.stab[stabcell][0] += 1
            thisrule.stab[stabcell][1] += tdif

    ##  now reward deviations from norm :
    tdif = 0.0
    for sr in range(paradat.stabrows):
        stabcell = int2bin(sr,paradat.subrules)
        if not(stabcell in thisrule.stab):  ##  fallback
            thisrule.stab[stabcell] = [1,0]
        else:
            td = thisrule.stab[stabcell][1]
            nc = thisrule.stab[stabcell][0]
            v = td/nc
            thisrule.stab[stabcell][1] = v
            tdif += abs(td)

    dbar = tdif / len(indices)
    if paradat.brevity == 0:
        return  [dbar]
    ##  otherwise append size penalty :
    totsize = 0
    for t in thisrule.testlist:
        totsize += treelib.size(t)

    return  [dbar,-totsize/len(thisrule.testlist)]
    



Varnames = ["zero","unit","pair","four","luck","Minus99","nine",
            "Zero","sub1","half","name","text"]

##  test blocks :
if 0 > 0:  ##  testing
    print(Varnames)
    ##  pseudo-context :
    paradat = Sack()
    paradat.subrules = 3
    paradat.taglist = Varnames
    paradat.thisrow = [0,1,2,4,7,-99,9,0,-1,0.5,"Hello World!","Pythonic"]
    paradat.cols = len(paradat.taglist)
    assert len(paradat.thisrow) == paradat.cols
    paradat.rows = 1
    paradat.allcols = list(range(paradat.cols))
    paradat.skipcols = set([0,8])  ##  forbidden
    paradat.goodcols = set(paradat.allcols) - paradat.skipcols
    paradat.strcols = [10,11]
    paradat.goodcols = paradat.goodcols - set(paradat.strcols)
    paradat.goodlist = sorted(paradat.goodcols)  ##  to save repeated re-listing
    ##  specimen string constants :
    paradat.strdict = {11: ["ello",'!',"zonk","lo","Hell","wo",'W'],
                       10: ["ic",'P','y',"zonq","xxzz",'@',"th"]}
    paradat.cols2use = sorted(paradat.goodlist + paradat.strcols)
    print(paradat.goodcols,paradat.goodlist,paradat.cols2use)

    paradat.dateline = time.asctime(time.localtime(time.time()))
    paradat.traindat = "made-up"
    paradat.rulemode = "tabular"
    paradat.targval = "zero"
    paradat.targcol = -1  ##  paradat.taglist.index(paradat.targval)
    paradat.catlist = [0,1]
    paradat.priorvec = [0.5,0.5]
    paradat.stabrows = 2**paradat.subrules

    for j in range(4):
        tx = treelib.maketree(paradat,random.randint(4,7))
        print(treelib.tree2str(tx,paradat))
        v = treelib.treeval(tx,paradat,paradat.thisrow)
        print(v)
    print()
    rx = makerule(paradat)
    outlines = ruleshow(paradat,rx)
    for rout in outlines:
        print(rout)
    sigcell = rule2sig(paradat,rx,paradat.thisrow)
    print(sigcell)
    
    
