"Python3 BEAGLE holdout-data regression-rule evaluation lister."

##  TREE :
##
##  The Regression Estimation Evaluator.
##  Copyright (c), 2016, R.S. Forsyth.
##  ancestral progenitor : Sept, 1980.
##  first version : 15/08/2015
##  last revision : 13/09/2016


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)


import  copy
import  math
import  os, sys
import  random, time

##  local libraries :
from globdefs import *
from glib import  numeric, numerize
from glib import  parafind, paraget
from glib import  paracopy
from glib import  parashow
from glib import  showobj
from glib import  openfile
from glib import  request
from glib import  pybyebye
from gridlib import  cols2use, matread
from gridlib import  matdump, vecline
from gridlib import  var2col
from gridlib import  bastatz, midvalue
from gridlib import  vecdist
from gridlib import  copygrid, distest
from gridlib import  matshow, precrec
from gridlib import  sckappa2
from gridlib import  hedgesg
##from gridlib import  corsums, corrval
import rulelib
from root import  regtest
from root import  showregs
from root import  targets


##  parameter info :
##  List of parameters that need 2b set, with defaults & lo,hi limits if appropriate :
Paralist = [
    ##  ("atomize",1,0,1),
    ("randseed",0,0,999999999),  ##  random seed, <2 means haphazard seed
    ("jobname","runster"),  ##  name to link related files
    ("casefold",1,0,1),  ##  whether to fold to lower case
    ("testdat",''),  ##  file for input of data
    ##  ("cols2use",''),  ##  variables to be used
    ("targval",''),  ##  target variable
    ("targcol",0),  ##  target column number [2b reset]
    ("idfield",''),  ##  varname for o/p label
    ("skipvars",''),  ##  variables to be excluded
    ("trigfunx",0,0,1),  ##  whether to allow trig functions (Cosi,Sine)
    ("listfile",''),  ##  main output file
    ("dumpfile",''),  ##  output file for subsampling results (?)
    ("rulefile",''),  ##  file for o/p of rules
    ("outpath",os.pardir+os.sep+"op"),  ##  default o/p drive/path
    ("outfile",os.pardir+os.sep+"op"+os.sep+"leaf.txt"),  ##  logfile 4checking
    ("zonk",0,0,1)  ##  just a sentinel
    ]
##  Each item above is a tuple: name,default,minvalue,maxvalue.
##  Contents of parafile will take precedence over these defaults,
##  but parameters not in Paralist won't be used even if in file.
##  0/1 used as True/False (lazy but easy!).
##  parafile syntax, line by line:
##  paraname <blank(s)> paraval


##  global vars :
Version = "The Regression Estimation Evaluator, v2.1"
##  Gold = 5.0**0.5*0.5+0.5
Mincases = 2
Dattail = "_dat2"
Dumptail = "_dump"
Listtail = "_test"
Ruletail = "_rule"
Maxcats = 64
Dustbin = "??!!"
Longname = 26  ##  length after which category names will be truncated
Longlab = 20  ##  length after which label fields will be truncated
Outpars = ["dateline","progname","id","testdat","targval"]
Goodsign = '+'
Poorsign = '-'
Star = '*'


##  Main program :
##
def main():
    "Main line."
    
    ##  First some odds & ends concerned with orientation :
    args = sys.argv[:]          ##  gets command-line arguments
    progname = args[0]
    dateline = time.asctime(time.localtime(time.time()))
    paraname, prepaths = parafind(args,dateline,Version)
    startime = time.time() ; clockon = time.clock()

    if paraname:
        parafile = openfile(paraname,prepaths,[".pf",".par",".txt",'.'],'r')
    else:
        parafile = None
    ##  print(parafile)
    if parafile == None:
        print("Cannot proceed without a parameter file!",NL,"Program halts.")
        return  -1

    ##  Read parameter settings :
    paradat = paraget(Paralist,parafile)
    ##  print(list(paradat.__dict__.items()),NL)
    parafile.close()

    ##  do some tidying of paradat :
    paradat.dateline = dateline
    paradat.progname = progname
    if hasattr(paradat,"casefold"):
        paradat.foldcase = paradat.casefold
    elif hasattr(paradat,"foldcase"):
        paradat.casefold = paradat.foldcase
    if not paradat.outpath.endswith(os.sep):
        paradat.outpath = paradat.outpath + os.sep
        ##  must have final separator character.
        if not os.path.isdir(paradat.outpath):  ##  try another outpath
            paradat.outpath = os.getcwd() + os.sep + "op" + os.sep
            if not os.path.isdir(paradat.outpath):
                os.makedirs(paradat.outpath)  ##  create o/p folder
            paradat.outfile = paradat.outpath + os.path.basename(paradat.outfile)

    ##  prepare o/p files :
    if paradat.listfile:
        listname = paradat.listfile
    else:
        listname = paradat.outpath + paradat.jobname + Listtail + ".txt"
    paradat.listfile = open(listname,'w',encoding="utf8")

    if paradat.dumpfile:
        df = paradat.dumpfile
    else:
        df = paradat.outpath + paradat.jobname + Dumptail + ".dat"
    paradat.dumpfile = open(df,'w',encoding="utf-8")

    ##  locate & read input files :
    tf = ''
    if paradat.testdat and os.path.isfile(paradat.testdat):
        print("data to be read from",paradat.testdat)
        r = getfile(paradat,paradat.testdat)
        tf = paradat.testdat
    else:
        tf = paradat.outpath + paradat.jobname + Dattail + ".dat"
        print("data to be sought on",tf)
        if os.path.isfile(tf):
            r = getfile(paradat,tf)
        else:
            raise RuntimeError("No test-data file found!")
    if r < 0:
        raise RuntimeError("Problem reading test-data file "+tf)
    ##  vars 2b skipped will have already affected paradat.goodcols.
    paradat.testdat = os.path.abspath(tf)  ##  record name if successful.
    ##  print(paradat.testdat)

    ##  next open rule file :
    if paradat.rulefile:
        rf = paradat.rulefile
    else:
        rf = paradat.outpath + paradat.jobname + Ruletail + ".txt"
    paradat.rulefile = open(rf,'r',encoding="utf-8")

    ##  print(paradat.datamat[0])
    ##  print(paradat.datamat[-1])

    ##  establish target values :
    if paradat.targval:
        targets(paradat)
        ##  should make .targvals & update .cols2use
    else:
        raise RuntimeError("No target expression (targval) given!!")
    tvalues = sorted(set(paradat.targvals))

    paradat.goodlist = sorted(paradat.goodcols)
    ##  print(paradat.goodlist)

    ##  identify individual target values :
    catlist = []
    catfrex = {}
    nonnums = 0
    for r in range(paradat.rows):
        tv = paradat.targvals[r]
        if not isinstance(tv,(int,float)):
            nonnums += 1
        if tv in catfrex:
            catfrex[tv] += 1
        else:
            catfrex[tv] = 1  ## first encounter
        if not (tv in catlist):
            catlist.append(tv)
    paradat.catlist = sorted(catlist)
    paradat.cats = len(paradat.catlist)
    paradat.catfrex = catfrex
    rootrows = paradat.rows ** 0.5
    if paradat.cats < 2:
        raise RuntimeError("Insufficient target variation to continue!!")
    elif nonnums > 0:
        raise RuntimeError(paradat.targval+" isn't numeric!!")
    elif paradat.cats < rootrows:
        print("Only",paradat.cats,"distinct target values, as follows :")
        print("Target values :",paradat.catlist)
        print("Are you sure you want to do regression rather than classification?")
    else:
        print(paradat.cats,"target values.")

    ##  compute basic stats :]
    paradat.targstat = bastatz(paradat.targvals)  ##  all target values
    showobj(paradat.targstat)

    ##  read from rule-file :
    datlines = paradat.rulefile.readlines()
    bestrule = rulelib.getrule(paradat,datlines)
    ##  print(bestrule.rows)
    ##  showobj(bestrule)
    ##  showobj(paradat)
    
    ##  compatibility issues :
    ##  assert paradat.targvar == bestrule.targvar
    ##  update some parameter values from input rule :
    parasave = paracopy(paradat,paradat.rulefile)  ##  save input settings
    paramods(paradat,bestrule)  ##  revises selected settings from bestrule
    paradat.rulefile.close()
    bout = rulelib.ruleshow(paradat,bestrule)
    for b in bout:
        print(b)
    print(bestrule.catlist)
    print(bestrule.rows)

    ##  prepare test phase :
    print(paradat.rows)
    ##  apply ruleset to test cases :
    allrows = list(range(paradat.rows))
    paradat.rows = bestrule.rows  ##  for cellsize
    results = regtest(paradat,bestrule,allrows,"holdout")
    nres = len(results)
    print("instances processed =",nres)
    paradat.nres = nres
    ##  summarize results :
    dustbins = hh = hits = 0
    icol = -1
    head = ["safeness","cellsize","filepos","id","stabcell","predval","trueval",
            "mark"]
    head = Htab.join(head)  ##  R-style headline
    print(head,file=paradat.dumpfile)  ##  tab-delimited header
    if paradat.idfield and (paradat.idfield in paradat.taglist):
        icol = paradat.taglist.index(paradat.idfield)
    for i,tr in enumerate(sorted(results,reverse=True)):
        ##  print(tr,file=paradat.dumpfile)
        dumpline(paradat,tr,icol)  ##  format readable by R read.delim()
        if tr[-1] == Goodsign:
            hits += 1
        if tr[1] < 10:  ##  small cellsize
            dustbins += 1
        else:
            if tr[-1] == Goodsign:
                hh += 1
    hitrate = 100*hits/nres
    print("overall hitrate % =",round(hitrate,4))
    if paradat.rulemode == "tabular":
        dustrate = 100*dustbins/nres
        print("percentage small-sample decisions =",round(dustrate,4))
        hr = 100*hh / max(1,nres-dustbins)
        print("hitrate % omitting small-samples =",round(hr,4))
    dumpname = paradat.dumpfile.name
    paradat.dumpfile.close()
    
    ##  o/p of results :
    ##  rows = paradat.rows ; paradat.rows = bestrule.rows
    nout = showregs(paradat,results,bestrule,"holdout")
    paradat.rows = parasave.rows  ##  restore
    ##  restore other changes ?
    print("cases processed =",nout)
    print(NL,file=paradat.listfile)  ##  insert blanks
    parashow(paradat,paradat.listfile)  ##  just show parasave ?
    listname = paradat.listfile.name
    paradat.listfile.close()
    ##  parashow(paradat,fo) ; fo.close()

    print("main o/p listing written onto",os.path.abspath(listname))
    print("R-readable results written to",os.path.abspath(dumpname))
    print(progname,"done on",time.asctime(time.localtime(time.time())))
    print("after",round(time.time()-startime,5),"seconds.")  ##  secs since kickoff

    pybyebye()
    return  0




##  DUMPLINE :
def dumpline (paradat,tr,icol):
    "puts test record onto .dumpfile in R-readable format."

    outline = [tr[0],tr[1]]  ##  safeness & cellsize
    filepos = tr[2]
    if icol > -1:
        ix = paradat.datamat[filepos][icol]
    else:
        ix = str(filepos)
    outline.append(filepos+1)  ##  file position for checking (e.g. in R)
    outline.append(ix)  ##  id field
    outline.append(tr[-3])  ##  stabsig
    outline += list(tr[-2])  ##  (pred,true)
    outline.append(tr[-1])  ##  mark
    ##  print(outline,file=paradat.dumpfile)
    outline = [str(item) for item in outline]
    lout = Htab.join(outline)
    print(lout,file=paradat.dumpfile)
    
    return


##  PARAMODS :
def paramods (paradat,bestrule):
    "revises setting that need 2b updated from bestrule info."

    paradat.priorvec = bestrule.priorvec
    ##  not prior probs but various stats passed over.
    paradat.rulemode = bestrule.rulemode
    paradat.subrules = len(bestrule.testlist)
    paradat.stabrows = bestrule.stabrows  ##  computed in getrule()
    paradat.targcuts = bestrule.catlist  ##  cutlist really
    if bestrule.catlist != paradat.catlist:
        ##  print(bestrule.catlist) ; print(paradat.catlist)
        paradat.catlist = bestrule.catlist
    ##  over-ride local stats :
    paradat.smalldif = bestrule.priorvec[-1]
    paradat.targmean = bestrule.priorvec[1]
    paradat.targstat.std1 = bestrule.priorvec[2]
    ##  dangerous ?!
    ##  paradat.rows = bestrule.rows
    
    return


##  SHOWLINE :
def showline (paradat,rank,strength,ix,caselab,predval,mark,trueval,cellsize,
              predvals):
    "displays single line on paradat.listfile."

    fm = "{0:4d}  {1:8.2f} {2:4d}  {3:20s} {4:>12s} ~ {5:12s} {6:4d}"
    fm = fm.replace('~',mark,1)
    pv = str(predval)[:12] ; tv = str(trueval)[:12]
    out1 = fm.format(rank,strength,ix,caselab,pv,tv,cellsize)
    out2 = ''
    if paradat.rulemode != "tabular":
        w = 10
    elif len(predvals) > 7:
        w = 6
    else:
        w = 7
    for pv in predvals:
        out2 = out2 + " {0:{1}.2f}".format(pv,w)
    print(out1+out2,file=paradat.listfile)


##  OUTSHOW :
def outshow (paradat,cmat,seen,fo):
    "shows confusion matrix."
    ##  adapt to use targcuts as quasi-cats ?
    
    ##  show confusion matrix :
    ptab,rtab = precrec(cmat)
    keylist = sorted(cmat.keys())  ##  pairs
    wide = max([len(str(k)) for k in keylist])
    wide = wide // 2
    rc = paradat.nres  ##  number of cases actually processed
    ##  matshow & precall expect pred,true order :
    matshow(fo,cmat,Longname)
    sk = sckappa2(cmat,paradat.catlist)  ##  keylist ??
    k2 = round(sk[0],4)
    ##  print(keylist,paradat.catlist)
    hits = sk[2]
    print("Kappa value = ",k2,file=fo)
    print("Precision (%) by category :",file=fo)
    for c in sorted(ptab.keys()):
        print(str(c)[:Longname].ljust(wide),
              round(ptab[c],4),file=fo)
    print("Recall (%) by category :",file=fo)
    for c in sorted(rtab.keys()):
        print(str(c)[:Longname].ljust(wide),
              round(rtab[c],4),file=fo)

    ##  display indicators :
    print("\ncases =",rc,file=fo)
    print("cases with unseen category labels =",rc-seen,file=fo)
    if seen > 0:
        if seen == 1:
            word = "case"
        else:
            word = "cases"
        if seen < rc:
            print("cases with known category labels =",seen,file=fo)
            print("[results below only apply to",seen,word+']',file=fo)
        print("hits = ",hits,file=fo)
        percent = 100 * hits / seen
        print("percent hits =",round(percent,2),file=fo)
        
    return


##  GETFILE :
def getfile (paradat,fi):
    "obtains data grid from file."

    paradat.datamat,paradat.taglist,paradat.tagdict = matread(fi,paradat)
    paradat.rows = len(paradat.datamat)
    print("data rows =",paradat.rows)
    paradat.cols = len(paradat.taglist)
    print("data cols =",paradat.cols)
    print("column names :")
    print(paradat.taglist)
    if paradat.rows < Mincases:  ##  not enough data
        print("Not enough data cases !")
        print("Program halts.")
        return  -1
    goodcols = cols2use(paradat)
    paradat.goodcols = goodcols
    paradat.goodlist = sorted(paradat.goodcols)
    if len(goodcols) < 1:
        print("No numeric variables found in ",paradat.datfile,'!')
        print("Program halts.")
        return  -2

    return  0  ##  success, paradat altered




##  Call main and attempt to honour exit status:
if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(1)
