##  Using RUNSTER R template, version of 12/09/2016 :
##  rule written by pear.py ;
##  derived from training data : c:\beagling\takeoff\airbomb_dat1.dat;
##  generated on creation date : Wed Sep 14 15:43:36 2016;
##  dumped on Thu Sep 15 11:02:03 2016.
##
beag_gold = 5.0 ** 0.5 * 0.5 + 0.5  ##  global

##  helper functions :
beag_bool = function (v) {
	##  ensures same bool/math treatment as in Beagle :
	return  ((v > 0) + 0)
	}
beag_exor = function (v1,v2) {
	##  exclusive or, as in Beagle :
	return ((v1>0) != (v2>0))
	}
beag_root = function (v) {
	##  safe square root :
	if (v >= 0.0)  return (sqrt(v))
	else  return (-sqrt(abs(v)))
	}
beag_slog = function (v) {
	##  safe natural logarithm :
	if (v < 0)  return (-log(1+abs(v)))
	else  return  (log(1+v))
	}

runster_stabprep = function () {
	##  sets up fallout table :

	stab = list()
	stab[['00']] = c(3, 0.6666666666666666, 0.4714045207910317)
	stab[['01']] = c(9, 0.8888888888888888, 0.31426968052735443)
	stab[['10']] = c(45, -0.7777777777777778, 0.46613726585340065)
	stab[['11']] = c(24, 0.7916666666666666, 0.40611643103370687)

	##  unpacks stab lines.

	return (stab)
	}  ##  stabprep ends.


runster_regrule = function (vals,stab) {
	##  input vals should be a 1-row dataframe with appropriate colnames.
	##  target : bombing - fighting.
	##  rule mode is tabular.

	rule = c() ; bins = c('0','1')
	catlist = c(-1, -0.07792207792207792, 0, 1)
	priorvec = c(77, -0.07792207792207792, 0.9225350979313268, 1)
	subrules = 2
	##  compute rule values :
	rule[1] = (max(vals$topspeed,(vals$headroom + vals$carrying)) > 401)
	rule[2] = (beag_bool(vals$wingspan) & beag_bool((beag_root(vals$wingspan) - 3.6864) - beag_root(vals$cannons)))

	p = 0 ; b = c()
	while (p < subrules) {
		p = p + 1  ##  early-r, late-py
		v = (rule[p]>0)  ##  omit if standard
		b = c(b,bins[v+1])   ##  omit if standard
		}
	##  standard mode :
	smalldif = priorvec[4]  ##  should work for both
	##  tabular mode :
	b = paste(b,collapse='')  ##  omit if standard
	cellvals = stab[[b]]  ##  omit if standard
	cellsize = cellvals[1]  ##  omit if standard
	predval = cellvals[2]  ##  omit if standard
	standev = cellvals[3]  ##  omit if standard

	return (list(cellcode=b,predval=predval,standev=standev,smalldif=smalldif,cellsize=cellsize))
	}
##  regression rule ends.


berries = function (datframe) {
	##  Bionically Evolved Regression Rule In Executable Software :

	stab = runster_stabprep()
	rows = dim(datframe)[1]
	options(stringsAsFactors=FALSE)
	if (rows < 1)  return (NULL)
	for (r in 1:rows) {
		regvals = runster_regrule(datframe[r,],stab)
		if (r == 1) outframe = data.frame(regvals)
		else {
			outframe = rbind(outframe,regvals)
			}
		##  outframe[j,] = c(datframe[r,],unlist(regvals))
		}
	temp = cbind(datframe,outframe)
	options(stringsAsFactors=default.stringsAsFactors())
	return (temp)
	}
##  returns dataframe with predicted values added to each record.
##  trueval may not be known, so no 'success' mark computed.

##  ending.
